/* FWIW, copyright 2005 Matt Neuburg 
mailto:matt@tidbits.com
http://www.tidbits.com/matt/
*/
/*
 Congratulations, you've found the only file of any importance.
 The idea here is to show how to overlay scriptability on an existing app.
 I started with an app that creates and manipulates persons and pairs thru a graphical interface,
 and then made it scriptable.
 All the scriptability stuff resides in this file as a set of categories.
 The example shows an element, a property, an enumeration, a couple of read-only properties,
 and a verb - the main things that give people trouble when they start adding scriptability.
 See text file "Testing Pairs" for a test bed.
 */

#if DEBUG
inline void MyLog(NSString *format, ...)
{
	va_list args;
	
	va_start(args, format);
	NSLogv(format, args);
	va_end(args);
}
#else
inline void MyLog(NSString *format, ...)
{return;}
#endif

#import "Scriptability.h"
#import "MyObject.h"
#import "Person.h"
#import "Pair.h"


@implementation NSObject (MNscriptability)

- (void) returnError:(int)n string:(NSString*)s {
	NSScriptCommand* c = [NSScriptCommand currentCommand];
	[c setScriptErrorNumber:n];
	if (s)
		[c setScriptErrorString:s];
}

@end

@implementation MyObject (MNscriptability)

// we are the app delegate

- (BOOL)application:(NSApplication *)sender delegateHandlesKey:(NSString *)key {
	MyLog(@"handles %@",key);
	if ([key isEqualToString: @"personsArray"])
		return YES;
	return NO;
}

// we have one element: "personsArray" (facade for self->persons)

- (unsigned int)countOfPersonsArray {
	MyLog(@"countofpersons");
	return [persons count]; 
}

// needed by "get every person"

- (Person *)objectInPersonsArrayAtIndex:(unsigned int)i {
	MyLog(@"personAtIndex %i", i);
	return [persons objectAtIndex: i];
}

// these seem to solve NSReceiverEvaluationScriptError messages

- (Person*)valueInPersonsArrayAtIndex:(unsigned int)i {
	MyLog(@"valueInPersonsAtIndex %i", i);
	if (![[NSScriptCommand currentCommand] isKindOfClass:[NSExistsCommand class]])
		if (i >= [persons count]) {
			[self returnError:errAENoSuchObject string:@"No such person."];
			return nil;
	}
	return [persons objectAtIndex: i];
}

- (Person*) valueInPersonsArrayWithName: (NSString*) name {
	MyLog(@"valueInPersonsWithName %@", name);
	int i, u = [persons count];
	for (i=0; i<u; i++)
		if ([[[persons objectAtIndex:i] name] caseInsensitiveCompare:name] == NSOrderedSame)
			return [persons objectAtIndex: i];
	if (![[NSScriptCommand currentCommand] isKindOfClass:[NSExistsCommand class]]) {
		[self returnError:errAENoSuchObject string:@"No such person."];
		return [[Person alloc] init];
	}
	return nil;
}

// error-checking utilities
// dictate rules of person creation, name assignment

- (BOOL) canGivePerson:(Person*)p name:(NSString*)name {
	if (!name || [name isEqualToString:@""]) {
		[self returnError:errOSACantAssign string:@"Can't give person empty name."];
		return NO;		
	}	
	if ([self existsPersonWithName: name]) {
		[self returnError:errOSACantAssign string:@"Can't give person same name as existing person."];
		return NO;
	}
	return YES;
}

- (void) scripterWantsToChangeName:(NSString*)n of:(Person*)p {
	if ([n isEqualToString: [p name]]) return; // nothing to do
	if (![self canGivePerson:p name:n]) return;
	[p setName: n];
	[self updateInterface:self];
}

- (BOOL) canCreatePerson:(Person*)p {
	if (![self canGivePerson:p name:[p name]]) return NO;
	if (! ([p gender] == MALE || [p gender] == FEMALE)) {
		[self returnError:errOSACantAssign string:@"A person must have a gender."];
		return NO;
	}
	if ([[NSScriptCommand currentCommand] scriptErrorNumber] < 0)
		return NO; // some other routine has declared this person invalid
	return YES;
}

// person creation
// I think both of these may be needed to satisfy KVC

- (void)insertObject:(Person *)p inPersonsArrayAtIndex:(unsigned int)index {
	MyLog(@"newPerson: %@", [p name]); 
	if (![self canCreatePerson:p]) return;
	[p setMaster: self];
	[persons insertObject:p atIndex:index];
	[self updateInterface:self];
}

- (void)insertInPersonsArray:(Person *)p {
	MyLog(@"newPerson2: %@", [p name]);
	if (![self canCreatePerson:p]) return;
	[p setMaster: self];
	[persons addObject:p];
	[self updateInterface:self];
}

// person deletion (prevent)
// I think both of these may be needed to satisfy KVC too

- (void)removeObjectFromPersonsArrayAtIndex:(unsigned int)index {
	MyLog(@"deletePerson");
	[self returnError:OSAMessageNotUnderstood string:nil];
}

-(void)removeFromPersonsArrayAtIndex:(unsigned int)index {
	MyLog(@"deletePerson2");
	[self returnError:OSAMessageNotUnderstood string:nil];
}

// called by Person to complete pair generation

- (void) scripterWantsToPair: (Person*) firstPerson with: (Person*) secondPerson;
{
	if ([firstPerson pair] || [secondPerson pair]) {
		[self returnError:errOSACantAssign string:@"Can't pair a person who is already paired."];
		return;
	}
	[self pair: firstPerson with: secondPerson];
}

@end

@implementation Person (MNscriptability)

// return a reference

- (NSScriptObjectSpecifier *)objectSpecifier {
	MyLog(@"personObjectSpecifier");
	NSScriptClassDescription* appDesc = (NSScriptClassDescription*)[NSApp classDescription];
	return [[[NSNameSpecifier alloc] initWithContainerClassDescription:appDesc containerSpecifier:nil key:@"personsArray" name:[self name]] autorelease];
}

// name property
// notice trick for postponing error-check during creation

- (NSString *)personName
{
	MyLog(@"getPersonName");
	return name; 
}
- (void)setPersonName:(NSString *)aName
{
    MyLog(@"setPersonName");
	if ([[NSScriptCommand currentCommand] isKindOfClass: [NSCreateCommand class]])
		[self setName:aName]; // postpone check, MyObject will handle it
	else
		[master scripterWantsToChangeName:aName of:self];
}

// gender property: can assign during creation but not afterwards

- (int)personGender
{
	MyLog(@"getPersonGender");
	return gender; 
}
- (void)setPersonGender:(int)aGender
{
    MyLog(@"setPersonGender");
	if ([[NSScriptCommand currentCommand] isKindOfClass: [NSCreateCommand class]])
		[self setGender:aGender]; 
	else
		[self returnError:errOSACantAssign string:@"Can't change person's gender."];
}

// pair command

- (void)scripterSaysPair:(NSScriptCommand*)command {
	MyLog(@"pair");
	Person* p1 = [command evaluatedReceivers];
	Person* p2 = [[command evaluatedArguments] valueForKey:@"otherPerson"];
	if (self != p1 || self == p2) {
		[self returnError:errOSACantAssign string:@"Invalid pairing."];
		return;
	}
	[master scripterWantsToPair:p1 with:p2];
}

// partner and paired properties
// must have both halves of a get/set even if the property is read-only

- (id) personPartner {
	MyLog(@"getPersonPartner");
	Pair* myPair = [self pair];
	if (!myPair)
		return [NSNull null]; // missing value
	return ([myPair person1] == self ? [myPair person2] : [myPair person1]);
}

- (void) setPersonPartner:(id)newPartner {
	MyLog(@"setPersonPartner");
	[self returnError:errOSACantAssign string:@"Partner property is read-only. To set a person's partner, pair the person with another person."];
}

- (BOOL) personPaired {
	MyLog(@"getPersonPaired");
	return ([self pair] != nil);
}

- (void) setPersonPaired:(BOOL)newPaired {
	MyLog(@"setPersonPaired");
	[self returnError:errOSACantAssign string:@"Paired property is read-only."];
}

@end
